import '../active_cases/active_cases.dart';
import '../configuration/configuration_service.dart';
import '../configuration/data_classes/configuration.dart';
import '../data_service/data_service.dart';
import '../log/log.dart';
import 'package:background_fetch/background_fetch.dart';

void backgroundFetchHeadlessTask(String taskId) async
{
  BackgroundFetch.finish(taskId);
  switch (taskId)
  {
    case "com.supersquad.heroes_app.autoSaveTask":
      DataService().performBackUp();
      break;

    case "flutter_background_fetch":
      BackgroundFetch.scheduleTask(TaskConfig(
          taskId: "com.supersquad.heroes_app.autoSaveTask",
          delay: 1000,
          periodic: false,
          forceAlarmManager: true,
          stopOnTerminate: false,
          enableHeadless: true)
      );
      break;
  }
}

class BackgroundFetchService
{
  //Singleton pattern
  static final _instance = BackgroundFetchService._internal();
  factory BackgroundFetchService() => _instance;
  BackgroundFetchService._internal()
  {
    ConfigurationService configurationService = ConfigurationService();
    if (configurationService.isLoaded)
      _minimumFetchIntervalInMinutes = configurationService.configuration.backUpFrequency.inMinutes;
    configurationService.addOnConfigurationChangedCallback(_onConfigurationChanged);
  }

  int _minimumFetchIntervalInMinutes = 15;
  int _backUpTaskDelay = 5000;
  int _loadRemoteActiveCasesTaskDelay = 5000;

  DateTime _lastTimeLoadedActiveCases = DateTime(2000);

  bool _isStarted = false;
  bool get isStarted => _isStarted;

  bool _isStarting = false;

  Future<void> start() async
  {
    if (_isStarted || _isStarting)
      return;

    _isStarting = true;

    await BackgroundFetch.registerHeadlessTask(backgroundFetchHeadlessTask);
    await _configure();
    BackgroundFetch.start().then((result){
      _isStarted = true;
      _isStarting = false;
    });
  }

  Future<void> _restart() async
  {
    if (_isStarted) {
      await BackgroundFetch.stop();
      _isStarted = false;
    }
    else
      await BackgroundFetch.registerHeadlessTask(backgroundFetchHeadlessTask);

    await _configure();
    await BackgroundFetch.start();

    _isStarted = true;
  }

  Future<int> _configure() async
  {
    try {
      final result = await BackgroundFetch.configure(
        BackgroundFetchConfig(
          minimumFetchInterval: _minimumFetchIntervalInMinutes,
          forceAlarmManager: false,
          stopOnTerminate: false,
          startOnBoot: true,
          enableHeadless: true,
          requiresBatteryNotLow: false,
          requiresCharging: false,
          requiresStorageNotLow: false,
          requiresDeviceIdle: false,
          requiredNetworkType: NetworkType.NONE,
        ),
        _onBackgroundFetch
      );
      Log.addToLog("Configured BackgroundFetch. Status = $result");
      return result;
    }
    catch (error) {
      Log.addToLog("Caught exception while configuring BackgroundFetch.");
      return BackgroundFetch.FETCH_RESULT_FAILED;
    }
  }

  //There are no awaits in here, despite this function being async.
  Future<void> _onBackgroundFetch(String taskId) async
  {
    switch (taskId)
    {
      case "com.supersquad.heroes_app.autoSaveTask":
        Log.addToLog("Performing automatic save task (asynchronously).");
        DataService().performBackUp();
        break;

      case "com.supersquad.heroes_app.loadSickDBTask":
        if (DateTime.now().difference(_lastTimeLoadedActiveCases) > ActiveCases().loadRemoteActiveCasesFrequency) {
          Log.addToLog("Performing loading of remote active cases task (asynchronously).");
          ActiveCases().loadRemote();
          _lastTimeLoadedActiveCases = DateTime.now();
        }
        break;

      case "flutter_background_fetch":
        BackgroundFetch.scheduleTask(TaskConfig(
            taskId: "com.supersquad.heroes_app.autoSaveTask",
            delay: _backUpTaskDelay,
            periodic: false,
            forceAlarmManager: true,
            stopOnTerminate: false,
            enableHeadless: true)
        );
        BackgroundFetch.scheduleTask(TaskConfig(
            taskId: "com.supersquad.heroes_app.loadSickDBTask",
            delay: _loadRemoteActiveCasesTaskDelay,
            periodic: false,
            forceAlarmManager: true,
            stopOnTerminate: false,
            enableHeadless: true)
        );
        break;
    }
    BackgroundFetch.finish(taskId);
  }

  void _onConfigurationChanged(Configuration newConfiguration)
  {
    final newBackUpFrequency = newConfiguration.backUpFrequency;
    if (newBackUpFrequency != null && newBackUpFrequency.inMinutes != _minimumFetchIntervalInMinutes)
    {
      Log.addToLog("backUpFrequency changed: " +
          _minimumFetchIntervalInMinutes.toString() + " minutes -> " +
          newBackUpFrequency.inMinutes.toString() + " minutes.");
      _minimumFetchIntervalInMinutes = newBackUpFrequency.inMinutes;
      _restart();
    }
  }
}