import '../active_cases/active_cases.dart';
import '../active_cases/data_classes/active_case.dart';
import '../configuration/configuration_service.dart';
import '../configuration/data_classes/configuration.dart';
import '../data_classes/exposure_data.dart';
import '../encounters/data_classes/encounter_entry.dart';
import '../encounters/encounters.dart';
import '../heroes_class.dart';
import '../log/log.dart';
import '../registration/registration.dart';
import 'package:intl/intl.dart';
import 'package:meta/meta.dart';

class DataService
{
  //Singleton pattern
  static final _instance = DataService._internal();
  factory DataService() => _instance;
  DataService._internal()
  {
    _encounters = Encounters();
    _activeCases = ActiveCases();

    ConfigurationService configurationService = ConfigurationService();
    if (configurationService.isLoaded) {
      _activeCases.loadRemoteActiveCasesFrequency = configurationService.configuration.loadRemoteActiveCasesFrequency;
      _encounters.encounterWindow = configurationService.configuration.encounterWindow;
      _backUpFrequency = configurationService.configuration.backUpFrequency;
      _encounterIsRelevantWindow = configurationService.configuration.encounterIsRelevantWindow;
    }
    configurationService.addOnConfigurationChangedCallback(_onConfigurationChangedCallback);
  }

  static final Duration encounterIsRelevantWindowDefault = Duration(days: 20);
  Duration _encounterIsRelevantWindow = encounterIsRelevantWindowDefault;

  bool _isStarted = false;
  bool get isStarted => _isStarted;

  Encounters _encounters;
  ActiveCases _activeCases;

  DateTime _lastTimePerformedBackUp = DateTime(2000);
  Duration _backUpFrequency = Duration(minutes: 15);

  Future<bool> start() async
  {
    if (isStarted)
      return true;

    await update();

    _isStarted = true;

    print("HEROES_FLUTTER_PLUGIN :: Data service started.");
    Log.addToLog("DataService started.");

    return _isStarted;
  }

  Future<bool> update() async
  {
    if (!_encounters.isLoaded)
      await _encounters.load(removeOlderThan: _encounterIsRelevantWindow);

    if (!_activeCases.isLoaded)
      await _activeCases.load(userID: Registration().userID, removeOlderThan: _encounterIsRelevantWindow);
    else
      await _activeCases.loadRemote(userID: Registration().userID);

    if (_encounters.isLoaded && _activeCases.isLoaded) {
      _checkAndAddNewActiveCases();
      print("HEROES_FLUTTER_PLUGIN :: Updated active cases.");
      Log.addToLog("DataService updated active cases.");
      return true;
    }

    return false;
  }

  void _checkAndAddNewActiveCases()
  {
    final remoteActiveCases = _activeCases.remoteActiveCases;
    final encountersMap = _encounters.encountersMap;

    final activeCasesToAdd = List<ActiveCase>();

    for (ActiveCase remoteActiveCase in remoteActiveCases)
      if (encountersMap.containsKey(remoteActiveCase.id))
        activeCasesToAdd.add(ActiveCase(id: remoteActiveCase.id, startDate: remoteActiveCase.startDate));

    _activeCases.addActiveCases(activeCasesToAdd);

    if (Heroes().onExposureDetected != null)
      Heroes().onExposureDetected(getTotalEncounterTime());
  }

  Future<bool> performBackUp() async
  {
    if (DateTime.now().difference(_lastTimePerformedBackUp) < _backUpFrequency)
      return false;

    bool encountersBackedUp = false;
    bool activeCasesBackedUp = false;

    if (_encounters.isLoaded)
      encountersBackedUp = await _encounters.saveEncountersToLocal();

    if (encountersBackedUp) {
      print("HEROES_FLUTTER_PLUGIN :: Performed encounters backup.");
      Log.addToLog("DataService performed encounters backup.");
    }

    if (_activeCases.isLoaded)
      activeCasesBackedUp = await _activeCases.saveActiveCasesToLocal();

    if (activeCasesBackedUp) {
      print("HEROES_FLUTTER_PLUGIN :: Performed active cases backup.");
      Log.addToLog("DataService performed active cases backup.");
    }

    final success = encountersBackedUp && activeCasesBackedUp;
    
    if (success)
      _lastTimePerformedBackUp = DateTime.now();

    return success;
  }

  ExposureData getTotalEncounterTime()
  {
    if (!_encounters.isLoaded || !_activeCases.isLoaded)
      return null;

    final activeCases = _activeCases.activeCases;
    final encountersMap = _encounters.encountersMap;

    final timeOfCalculation = DateTime.now();
    Duration totalDuration = Duration();

    for (ActiveCase activeCase in activeCases)
      if (encountersMap.containsKey(activeCase.id))
      {
        final encountersWithOneDevice = encountersMap[activeCase.id];

        for (EncounterEntry encounterEntry in encountersWithOneDevice)
          totalDuration += encounterEntry.lastTime.difference(encounterEntry.fistTime);
      }

    return ExposureData(totalDuration, timeOfCalculation);
  }

  void startEncounter(String id, DateTime dateTime)
  {
    final match = _activeCases.activeCases.firstWhere((activeCase) => activeCase.id == id, orElse: () => null);
    if (match == null)
      _encounters.startEncounter(id, dateTime);
    else {
      print("HEROES_FLUTTER_PLUGIN :: DATA_SERVICE :: startEncounter() :: " + id + " is in active cases list. Discarding this encounter.");
      Log.addToLog("Encountered " + id.substring(0, 6) + " but it's an active case, so we ignored it.");
    }
  }

  void finishEncounter(String id)
  {
    _encounters.finishEncounter(id);
  }

  void finishAllEncounters()
  {
    _encounters.finishAllEncounters();
  }

  void _onConfigurationChangedCallback(Configuration newConfiguration)
  {
    bool FLAG_RELOAD_REMOTE_ACTIVE_CASES = false;
    bool FLAG_REMOVE_OLD_ENCOUNTERS = false;
    bool FLAG_REMOVE_OLD_ACTIVE_CASES = false;
    bool FLAG_PERFORM_BACKUP = false;

    final newEncounterIsRelevantWindow = newConfiguration.encounterIsRelevantWindow;
    if (_encounterIsRelevantWindow != newEncounterIsRelevantWindow)
    {
      Log.addToLog("encounterIsRelevantWindow changed: " +
          _encounterIsRelevantWindow.inDays.toString() + " days -> " +
          newEncounterIsRelevantWindow.inDays.toString() + " days.");

      if (newEncounterIsRelevantWindow < _encounterIsRelevantWindow) {
        FLAG_REMOVE_OLD_ACTIVE_CASES = true;
        FLAG_REMOVE_OLD_ENCOUNTERS = true;
      }
    }

    final newLoadRemoteActiveCasesFrequency = newConfiguration.loadRemoteActiveCasesFrequency;
    if (_activeCases.loadRemoteActiveCasesFrequency != newLoadRemoteActiveCasesFrequency)
    {
      Log.addToLog("loadRemoteActiveCasesFrequency changed: " +
          _activeCases.loadRemoteActiveCasesFrequency.inHours.toString() + " hours -> " +
          newLoadRemoteActiveCasesFrequency.inHours.toString() + " hours.");

      _activeCases.loadRemoteActiveCasesFrequency = newConfiguration.loadRemoteActiveCasesFrequency;

      if (DateTime.now().difference(_activeCases.lastTimeLoadedRemoteActiveCases) > _activeCases.loadRemoteActiveCasesFrequency) {
        FLAG_RELOAD_REMOTE_ACTIVE_CASES = true;
      }
    }

    final newEncounterWindow = newConfiguration.encounterWindow;
    if (_encounters.encounterWindow != newEncounterWindow)
    {
      Log.addToLog("encounterWindow changed: " +
          _encounters.encounterWindow.inMinutes.toString() + " minutes -> " +
          newEncounterWindow.inMinutes.toString() + " minutes.");

      _encounters.encounterWindow = newEncounterWindow;
    }

    final newBackUpFrequency = newConfiguration.backUpFrequency;
    if (newBackUpFrequency.inMinutes != _backUpFrequency.inMinutes)
    {
      Log.addToLog("backUpFrequency changed: " +
          _backUpFrequency.inMinutes.toString() + " minutes -> " +
          newBackUpFrequency.inMinutes.toString() + " minutes.");

      if (DateTime.now().difference(_lastTimePerformedBackUp) > _backUpFrequency) {
        FLAG_PERFORM_BACKUP = true;
      }
    }

    _applyConfigurations(
      FLAG_RECALCULATE_ENCOUNTERS_WITH_ACTIVE_CASES: FLAG_RELOAD_REMOTE_ACTIVE_CASES,
      FLAG_REMOVE_OLD_ENCOUNTERS: FLAG_REMOVE_OLD_ENCOUNTERS,
      FLAG_REMOVE_OLD_ACTIVE_CASES: FLAG_REMOVE_OLD_ACTIVE_CASES,
      FLAG_PERFORM_BACKUP: FLAG_PERFORM_BACKUP,
    );
  }

  Future<void> _applyConfigurations({
    @required bool FLAG_RECALCULATE_ENCOUNTERS_WITH_ACTIVE_CASES,
    @required bool FLAG_REMOVE_OLD_ENCOUNTERS,
    @required bool FLAG_REMOVE_OLD_ACTIVE_CASES,
    @required bool FLAG_PERFORM_BACKUP,}) async
  {
    bool FLAG_RELOADED_REMOTE_ACTIVE_CASES = false;
    bool FLAG_REMOVED_OLD_ACTIVE_CASES = false;
    bool FLAG_RELOADED_ENCOUNTERS = true;
    bool FLAG_REMOVED_OLD_ENCOUNTERS = false;

    if (FLAG_RECALCULATE_ENCOUNTERS_WITH_ACTIVE_CASES)
    {
      if (_activeCases.isLoaded)
        _activeCases.saveActiveCasesToLocal();
      else {
        _activeCases.load(removeOlderThan: _encounterIsRelevantWindow);
        FLAG_RELOADED_REMOTE_ACTIVE_CASES = true;
        FLAG_REMOVED_OLD_ACTIVE_CASES = true;
      }

      if (_encounters.isLoaded)
        _encounters.saveEncountersToLocal();
      else {
        _encounters.load(removeOlderThan: _encounterIsRelevantWindow);
        FLAG_RELOADED_ENCOUNTERS = true;
        FLAG_REMOVED_OLD_ENCOUNTERS = true;
      }

      _checkAndAddNewActiveCases();
    }

    if (FLAG_REMOVE_OLD_ENCOUNTERS && !FLAG_REMOVED_OLD_ENCOUNTERS)
      _encounters.removeEncountersOlderThan(_encounterIsRelevantWindow);

    if (FLAG_REMOVE_OLD_ACTIVE_CASES && !FLAG_REMOVED_OLD_ACTIVE_CASES)
      _activeCases.removeActiveCasesOlderThan(_encounterIsRelevantWindow);

    if (FLAG_PERFORM_BACKUP)
      if (_encounters.isLoaded)
        _encounters.saveEncountersToLocal();
      if (_activeCases.isLoaded)
        _activeCases.saveActiveCasesToLocal();
  }

  bool isSick(String deviceId)
  {
    final match = ActiveCases().activeCases.firstWhere((ac) => ac.id == deviceId, orElse: () => null);
    return match != null;
  }

  List<List<EncounterEntry>> getAllEncountersFromDB()
  {
    List<List<EncounterEntry>> result = new List<List<EncounterEntry>>();
    Encounters().encountersMap.forEach((k, v) => result.add(v));
    return result;
  }

  String getEncountersFormatted(String id)
  {
    String entry = "";
    if (Encounters().encountersMap.containsKey(id)) {
      List<EncounterEntry> encounters = Encounters().encountersMap[id];
      for (int j = 0; j < encounters.length; j++)
        entry += "Encounter: \nStartDate=" +
            DateFormat("yyyy-MM-dd HH:mm:ss").format(encounters[j].fistTime) +
            ", endDate=" +
            DateFormat("yyyy-MM-dd HH:mm:ss").format(encounters[j].lastTime) +
            "\n\n";
    }
    return entry;
  }

  List<EncounterEntry> getCurrentEncounters()
  {
    final result = List<EncounterEntry>();
    for (String currentEncounterID in Encounters().currentEncounters)
      if (Encounters().encountersMap.containsKey(currentEncounterID))
        result.add(Encounters().encountersMap[currentEncounterID].last);
    return result;
  }
}