import 'data_classes/encounter_entry.dart';
import 'repositories/encounters_repository.dart';
import '../app_clock_service/app_clock.dart';
import '../encounters/data_classes/encounter_position.dart';
import '../gps_service/gps_service.dart';

class Encounters
{
  static Encounters _instance = Encounters._internal();
  factory Encounters() => _instance;
  Encounters._internal()
  {
    _repository = EncountersRepository();
  }

  bool _isLoaded = false;
  bool get isLoaded => _isLoaded;

  EncountersRepository _repository;

  Map<String, List<EncounterEntry>> get encountersMap => _mapDB;
  Map<String, List<EncounterEntry>> _mapDB = new Map<String, List<EncounterEntry>>();

  List<String> get currentEncounters => _currentEncounters;
  List<String> _currentEncounters = List<String>();

  Duration get encounterWindow => _encounterWindow;
  set encounterWindow(Duration value) => _encounterWindow = value;
  Duration _encounterWindow = Duration(minutes: 5);

  Future<void> load({Duration removeOlderThan = null}) async
  {
    await _loadEncountersFromLocal();

    _isLoaded = true;

    if (removeOlderThan != null)
      removeEncountersOlderThan(removeOlderThan);
  }

  Future<bool> _loadEncountersFromLocal() async
  {
    final result = await _repository.getEncounters();
    if (result != null) {
      _mapDB = result;
      print("HEROES_FLUTTER_PLUGIN :: Loaded " + result.length.toString() + " encounters from local database into encounters map.");
      return true;
    }
    else {
      print("HEROES_FLUTTER_PLUGIN :: Could not load encounters from local database into encounters map.");
    }
    return false;
  }

  void removeEncountersOlderThan(Duration duration)
  {
    final entries = _mapDB.entries.toList();

    for (MapEntry<String, List<EncounterEntry>> entry in entries)
    {
      for (int i=entry.value.length-1; i>=0; i--) {
        if (DateTime.now().difference(entry.value[i].lastTime) >= duration)
          entry.value.removeAt(i);
      }
      if (entry.value.length <= 0)
        _mapDB.remove(entry.key);
    }
    print("HEROES_FLUTTER_PLUGIN :: Removed old (over " + duration.inDays.toString() + " days) encounters from local map, now containing " + _mapDB.length.toString() + " encounters.");

    saveEncountersToLocal();
  }

  Future<bool> saveEncountersToLocal() async {
    final result = await _repository.saveEncounters(_mapDB);
    if (result)
      print("HEROES_FLUTTER_PLUGIN :: Saved " + _mapDB.length.toString() + " encounters from encounters map to local database.");
    else
      print("HEROES_FLUTTER_PLUGIN :: Could not save encounters from encounters map to local database.");
    return result;
  }

  void startEncounter(String id, DateTime dateTime)
  {
    //Check that this is not an old message
    if (DateTime.now().difference(dateTime) <= _encounterWindow)
    {
      EncounterPosition position = new EncounterPosition(lat: GPSService().latitude, long: GPSService().longitude);
      _addToCurrentEncountersList(id);

      //Check whether the encountered device already exists in DB
      if (_mapDB.containsKey(id))
      {
        if (dateTime.difference(_mapDB[id].last.lastTime) <= _encounterWindow)
        {
          _mapDB[id].last.lastTime = dateTime;
          _mapDB[id].last.positions.add(position);
        }
        else
        {
          List<EncounterPosition> positions = new List<EncounterPosition>();
          positions.add(position);
          _mapDB[id].add(new EncounterEntry(
              theirId: id,
              fistTime: dateTime,
              lastTime: dateTime,
              positions: positions));
        }
      }
      else
      {
        //If it doesn't exist, register this device with this encounter in DB
        List<EncounterPosition> positions = new List<EncounterPosition>();
        positions.add(position);
        EncounterEntry encounter = new EncounterEntry(
            theirId: id,
            fistTime: dateTime,
            lastTime: dateTime,
            positions: positions);
        List<EncounterEntry> deviceEncounters = new List<EncounterEntry>();
        deviceEncounters.add(encounter);
        final mapEntry = MapEntry(id, deviceEncounters);
        final mapEntryList = List<MapEntry<String, List<EncounterEntry>>>();
        mapEntryList.add(mapEntry);
        _mapDB.addEntries(mapEntryList);
      }
    }
    else
    {
      print("HEROES_FLUTTER_PLUGIN :: Discard old message");
    }
  }

  void finishEncounter(String id) {
    if (_mapDB.containsKey(id)) {
      _removeFromCurrentEncountersList(id);
      _mapDB[id].last.lastTime = DateTime.now();
    }
  }

  void finishAllEncounters() {
    for (int i = 0; i < _currentEncounters.length; i++) {
      if (_mapDB.containsKey(_currentEncounters[i]))
        _mapDB[_currentEncounters[i]].last.lastTime = DateTime.now();
    }
    _currentEncounters.clear();
  }

  void _addToCurrentEncountersList(String id) {
    if (!_currentEncounters.contains(id)) _currentEncounters.add(id);
  }

  void _removeFromCurrentEncountersList(String id) {
    if (_currentEncounters.contains(id)) {
      for (int i = 0; i < _currentEncounters.length; i++) {
        if (_currentEncounters[i] == id) _currentEncounters.removeAt(i);
      }
    }
  }
}