import 'app_clock_service/app_clock.dart';
import 'background_fetch/background_fetch.dart';
import 'configuration/configuration_service.dart';
import 'data_classes/exposure_data.dart';
import 'data_service/data_service.dart';
import 'detection_service/detection_service.dart';
import 'gps_service/gps_service.dart';
import 'log/log.dart';
import 'registration/registration.dart';

class Heroes
{
  //Singleton pattern
  static final _instance = Heroes._internal();
  factory Heroes() => _instance;
  Heroes._internal() {
    _configuration = ConfigurationService();
    _registration = Registration();
    _dataService = DataService();
    _gpsService = GPSService();
    _detectionService = DetectionService();
    _backgroundFetchService = BackgroundFetchService();
  }

  Function(ExposureData) onExposureDetected;

  ConfigurationService _configuration;
  Registration _registration;
  DataService _dataService;
  GPSService _gpsService;
  DetectionService _detectionService;
  BackgroundFetchService _backgroundFetchService;

  Future<bool> get userIsRegistered async => await _registration.isRegistered;

  /// Registers a user.
  /// Returns [bool] success.
  Future<bool> registerUser(String invitationCode) async
  {
    if (invitationCode == null || invitationCode == "")
      return false;

    final result = await _registration.registerUser(inviteCode: invitationCode);

    if (result == null)
      return false;

    if (result.success)
      print("HEROES_FLUTTER_PLUGIN :: Successfully registered user.");
    else
      print("HEROES_FLUTTER_PLUGIN :: User was not registerd.");

    return result.success;
  }

  String _APIkey = "";
  Future<void> setAPIkey(String value) async
  {
    _APIkey = value;

    if (!(_configuration.isLoaded))
      await _configuration.loadConfiguration(APIkey: _APIkey);
  }

  String get userID => _registration.userID;
  List<String> get invitationCodes => _registration.invitationCodes;

  bool _isStarted = false;
  bool get isStarted => _isStarted;

  /// Returns [bool] success.
  Future<bool> start({String APIkey = ""}) async
  {
    if (isStarted)
      return true;

    if (APIkey != null && APIkey != "")
      _APIkey = APIkey;
    else if (_APIkey == "")
      return false;

    Log.addToLog("Starting Heroes with API key: $_APIkey");

    if (!(_configuration.isLoaded))
      await _configuration.loadConfiguration(APIkey: _APIkey);

    AppClock(); //Initialize AppClock's static values.

    if (!(await userIsRegistered)) {
      Log.addToLog("User is not registered.");
      return false;
    }
    else
      Log.addToLog("User is registered. ID = " + _registration.userID.substring(0, 6) + ".");

    if (!(_dataService.isStarted))
      await _dataService.start();

    await _gpsService.init();

    if (!_detectionService.isStarted)
      _detectionService.start(); //Waiting here takes +20s

    if (!_backgroundFetchService.isStarted)
      await _backgroundFetchService.start();

    print("HEROES_FLUTTER_PLUGIN :: Started.");
    Log.addToLog("Heroes started.");

    _isStarted = true;
    return true;
  }

  Future<void> stop() async
  {
    _dataService.finishAllEncounters();
    _detectionService.stop(); //We don't wait here.
    await _dataService.performBackUp();
    return null;
  }

  Future<bool> performBackUp() async
  {
    return await _dataService.performBackUp();
  }

  ExposureData getExposureData()
  {
    return _dataService.getTotalEncounterTime();
  }

  void setLogCallback(Function function)
  {
    Log.setAddToLogCallback(function);
  }

  @override
  String toString() => "Heroes(onExposureDetected: " +
      (onExposureDetected == null ? "null" : "not null") + ")";
}