import '../app_clock_service/app_clock.dart';
import '../configuration/configuration_service.dart';
import '../configuration/data_classes/configuration.dart';
import '../data_service/data_service.dart';
import '../log/log.dart';
import '../registration/registration.dart';
import 'dart:async';
import 'dart:convert';
import 'package:nearby_connections/nearby_connections.dart';

class NearbyConnectionsService
{
  factory NearbyConnectionsService() => _instance;

  NearbyConnectionsService._internal()
  {
    print("HEROES_FLUTTER_PLUGIN :: NearbyConnectionsService :: Constructor.");
    ConfigurationService configurationService = ConfigurationService();
    if (configurationService.isLoaded) {
      _serviceID = configurationService.configuration.serviceID;
      _maxTimeWithoutEventsForReboot = configurationService.configuration.maxTimeWithoutEventsForReboot;
    }
    configurationService.addOnConfigurationChangedCallback(_onConfigurationChangedCallback);
  }

  String _serviceID = "";

  DateTime _startTime;
  DateTime _stopTime;
  DateTime _lastNearbyEventTime;

  Duration _maxTimeWithoutEventsForReboot = Duration(minutes: 3);

  bool _lock;

  static final _instance = NearbyConnectionsService._internal();
  static final Strategy strategy = Strategy.P2P_CLUSTER;

  Registration _registration = Registration();
  DataService _dataService = DataService();

  Function _onDetectedCallback = (dynamic params) => print("HEROES_FLUTTER_PLUGIN :: NearbyConnectionsService :: onDetectedCallback function is not assigned.");
  set onDetectedCallback(Function function) {
    if (function != null)
      _onDetectedCallback = function;
  }

  Function _onLostCallback = (dynamic params) => print("HEROES_FLUTTER_PLUGIN :: NearbyConnectionsService :: onLostCallback function is not assigned.");
  set onLostCallback(Function function) {
    if (function != null)
      _onLostCallback = function;
  }

  Future<void> init() async {
    print("HEROES_FLUTTER_PLUGIN :: StartConnections");
    Log.addToLog("Initiating NearbyConnectionsService");
    _lock = false;
    _lastNearbyEventTime = DateTime.now();

    await stop();
    await detectDevices();
    AppClock().startIsolate(reboot);
  }

  Future<void> detectDevices() async {
    await _startAdvertising();
    await _startDiscovery();
    _startTime = DateTime.now();
    Duration timeSinceLastStop = _startTime.difference(_stopTime);
    Log.addToLog("Start :: " + timeSinceLastStop.inSeconds.toString() + " sg");
    Log.addToLog("Start Advertising/Discovery with id " + _getProperlyEncodedString(_registration.userID).substring(0, 6));
  }

  Future<void> stop() async {
    await _stopAdvertising();
    await _stopDiscovery();
    Log.addToLog("Stopped Advertising/Discovery");
    _stopTime = DateTime.now();
  }

  Future<void> reboot() async
  {
    if(!_lock) // avoid isolate thread callback getting in while previous one is still running
    {
      _lock = true;

      bool doReboot = false;

      if (DateTime.now().difference(_lastNearbyEventTime) >= _maxTimeWithoutEventsForReboot){
        _lastNearbyEventTime = DateTime.now(); //<= if not updated every time the reboot is forced it will enter here always if no previous OnDetect/OnLost has been detected, which is the whole point of this!!
        doReboot = true;
      }

      if (doReboot) {
        print("HEROES_FLUTTER_PLUGIN :: StartReboot");
        Log.addToLog("Starting NearbyConnectionsService reboot.");
        _dataService.finishAllEncounters();
        await stop();
        await detectDevices();
      }
      _lock = false;
    }
  }

  Future<void> _startAdvertising() async {
    try {
      final start = DateTime.now();
      await Nearby().startAdvertising(
          _getProperlyEncodedString(_registration.userID), strategy,
          onConnectionInitiated: onConnectionInit,
          onConnectionResult: (id, status) {},
          onDisconnected: (id) {},
          serviceId: _serviceID);
      print("HEROES_FLUTTER_PLUGIN :: StartAdvertising");
      Log.addToLog("Started advertising with " + _getProperlyEncodedString(_registration.userID).substring(0, 6) +
          " (Nearby().startAdvertising() took " + DateTime.now().difference(start).inMilliseconds.toString() + " ms).");
    } catch (exception) {
      print("HEROES_FLUTTER_PLUGIN :: " + exception.toString());
      Log.addToLog("Caught exception in await Nearby().startAdvertising();");
    }
    return;
  }

  Future<void> _startDiscovery() async {
    try {
      final start = DateTime.now();
      await Nearby().startDiscovery(
          _getProperlyEncodedString(_registration.userID), strategy,
          onEndpointFound: (id, name, serviceId) {
            if (_stringIsValid(name)) {
              print("HEROES_FLUTTER_PLUGIN :: OnFound " + name.substring(0, 6));
              Log.addToLog("Found EndPoint " + id + " with UUID " + name.substring(0, 6));
              _lastNearbyEventTime = DateTime.now();
              _onEndPointFound(id, name);
            }
            else {
              final result = _tryGettingProperStringFrom(name);
              if (result != null) {
                print("HEROES_FLUTTER_PLUGIN :: OnFound");
                Log.addToLog("Found EndPoint " + id + " with UUID " + result.substring(0, 6));
                _lastNearbyEventTime = DateTime.now();
                _onEndPointFound(id, result);
              }
              else
                Log.addToLog("Found EndPoint with corrupted name and discarded it.");
            }
          },
          onEndpointLost: (id) {
            _lastNearbyEventTime = DateTime.now();
            print("HEROES_FLUTTER_PLUGIN :: OnLost");
            Log.addToLog("Lost EndPoint " + id);
            _onEndPointLost(id);
          },
          serviceId: _serviceID);
      print("HEROES_FLUTTER_PLUGIN :: StartDiscovery");
      Log.addToLog("Started discovery with " + _getProperlyEncodedString(_registration.userID).substring(0, 6) +
          " (Nearby().startDiscovery() took " + DateTime.now().difference(start).inMilliseconds.toString() + " ms).");
    }
    catch (e) {
      print("HEROES_FLUTTER_PLUGIN :: " + e.toString());
      Log.addToLog("Caught exception in await Nearby().startDiscovery();");
    }

    return;
  }

  void onConnectionInit(String id, ConnectionInfo info) {
    //TODO: IMPLEMENT IF IS NEDEED
  }

  Future<void> _stopAdvertising() async {
    final start = DateTime.now();
    await Nearby().stopAdvertising();
    final duration = DateTime.now().difference(start);
    Log.addToLog("Nearby().stopAdvertising() took " + duration.inMilliseconds.toString() + " ms.");
    return null;
  }

  Future<void> _stopDiscovery() async {
    final start = DateTime.now();
    await Nearby().stopDiscovery();
    final duration = DateTime.now().difference(start);
    Log.addToLog("Nearby().stopDiscovery() took " + duration.inMilliseconds.toString() + " ms.");
    return null;
  }

  //SHOULD BE ON UTILS
  final _validCharactersDefault = RegExp(r'^[a-zA-Z0-9]+$');

  bool _stringIsValid(String string, {RegExp validCharacters}) {
    if (validCharacters == null) validCharacters = _validCharactersDefault;

    final isValid = validCharacters.hasMatch(string);
    return isValid;
  }

  String _getProperlyEncodedString(String string, {RegExp validCharacters}) {
    if (validCharacters == null) validCharacters = _validCharactersDefault;

    if (validCharacters.hasMatch(string))
      return string;
    else {
      final idAsUTF16 = string.codeUnits;
      final id = String.fromCharCodes(idAsUTF16);
      if (validCharacters.hasMatch(id))
        return id;
      else {
        final idAsUTF8 = utf8.encode(string);
        final id = utf8.decode(idAsUTF8, allowMalformed: true);
        if (validCharacters.hasMatch(id)) return id;
      }
    }
    return string;
  }

  String _tryGettingProperStringFrom(String string) {
    final validCharacters = _validCharactersDefault;
    if (validCharacters.hasMatch(string))
      return string;
    else {
      final idAsUTF16 = string.codeUnits;
      final id = String.fromCharCodes(idAsUTF16);
      if (validCharacters.hasMatch(id))
        return id;
      else {
        final idAsUTF8 = utf8.encode(string);
        final id = utf8.decode(idAsUTF8, allowMalformed: true);
        if (validCharacters.hasMatch(id)) return id;
      }
    }
    return null;
  }

  void _onEndPointFound(String id, String name)
  {
    _onDetectedCallback({"id": id, "name": name});
  }

  void _onEndPointLost(String id)
  {
    _onLostCallback({"id": id});
  }

  void _onConfigurationChangedCallback(Configuration newConfiguration)
  {
    final newServiceID = newConfiguration.serviceID;
    if (_serviceID != newServiceID)
    {
      Log.addToLog("serviceID changed: \"" + _serviceID + "\" -> \"" + newServiceID + "\"");
      _serviceID = newServiceID;
    }

    final newMaxTimeWithoutEventsForReboot = newConfiguration.maxTimeWithoutEventsForReboot;
    if (_maxTimeWithoutEventsForReboot.inSeconds != newMaxTimeWithoutEventsForReboot.inSeconds)
    {
      Log.addToLog("maxTimeWithoutEventsForReboot changed: " +
          _maxTimeWithoutEventsForReboot.inSeconds.toString() + " seconds -> " +
          newMaxTimeWithoutEventsForReboot.inSeconds.toString() + " seconds.");
      _maxTimeWithoutEventsForReboot = newMaxTimeWithoutEventsForReboot;
    }
  }
}
